import 'dart:math';

import '../../complex_analysis/complex.dart';
import '../../discrete_mathematics/general_algebraic_systems/number/double.dart';
import '../../discrete_mathematics/general_algebraic_systems/number/integer.dart';
import 'base/equation_base.dart';
import 'exceptions/equation_exception.dart';
import 'quadratic_equation.dart';

/// Class for work with cubic equations
class CubicEquation extends EquationBase {
  /// Accepts coefficients [a], [b], [c] and [d]
  ///
  /// All values shouldn't be equal to zero. Otherwise the [EquationException] is thrown.
  CubicEquation({this.a = 1, this.b = 0, this.c = 0, this.d = 0}) {
    if (a == 0) {
      throw EquationException('Coefficient a shouldn\'t be equal to zero!');
    }
  }

  /// [a] coefficient in cubic equation
  num a;

  /// [b] coefficient in cubic equation
  num b;

  /// [c] coefficient in cubic equation
  num c;

  /// [d] coefficient in cubic equation
  num d;

  /// Calculates [p] in canonical cubic expression
  num get p => (c / a) - (pow(b, 2) / (3 * pow(a, 2)));

  /// Calculates [q] in canonical cubic expression
  num get q {
    final firstExpression = (2 * pow(b, 3)) / (27 * pow(a, 3));
    final secondExpression = (b * c) / (3 * pow(a, 2));
    final thirdExpression = d / a;
    return firstExpression - secondExpression + thirdExpression;
  }

  @override
  Map<String, Complex> calculate() {
    final result = <String, Complex>{};
    final dis = discriminant();

    if (d != 0) {
      final possibleX = Integer(d).factorizate();
      possibleX.add(1);

      for (var item in possibleX) {
        for (var i = 1; i <= 2; i++) {
          if (evaluateForZero(pow(-1, i) * item)) {
            result['x1'] = Complex(re: pow(-1, i) * item);
          }
        }
      }
    } else {
      result['x1'] = Complex();
    }

    if (result.isEmpty) {
      final alpha =
          Double(-(q / 2) + sqrt(dis)).preciseTo(2).rootOf(3).toDouble();
      final beta =
          Double(-(q / 2) - sqrt(dis)).preciseTo(2).rootOf(3).toDouble();
      final z = alpha + beta;
      final x = z - b / (3 * a);
      result['x1'] = Complex(re: x);
    }

    final tmpB = result['x1'] * a + b;
    final tmpC = result['x1'] * tmpB + c;

    final quadratic =
        QuadraticEquation(a: a, b: tmpB.toReal(), c: tmpC.toReal());
    final quadResult = quadratic.calculate();
    for (var i = 1; i <= quadResult.length; i++) {
      result['x${i + 1}'] = quadResult['x$i'];
    }

    return result;
  }

  @override
  num discriminant() => (pow(q, 2) / 4) + (pow(p, 3) / 27);

  @override
  bool evaluateForZero(num x) =>
      (a * pow(x, 3) + b * pow(x, 2) + c * x + d) == 0;

  @override
  String toString() => '${a}x^3 + ${b}x^2 + ${c}x + $d';
}
